from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.image_converter import ImageConverter
from keras_hub.src.models.siglip.siglip_backbone import SigLIPBackbone


@keras_hub_export("keras_hub.layers.SigLIPImageConverter")
class SigLIPImageConverter(ImageConverter):
    backbone_cls = SigLIPBackbone



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.modeling.masked_lm_head import MaskedLMHead
from keras_hub.src.models.distil_bert.distil_bert_backbone import (
    DistilBertBackbone,
)
from keras_hub.src.models.distil_bert.distil_bert_backbone import (
    distilbert_kernel_initializer,
)
from keras_hub.src.models.distil_bert.distil_bert_masked_lm_preprocessor import (  # noqa: E501
    DistilBertMaskedLMPreprocessor,
)
from keras_hub.src.models.masked_lm import MaskedLM


@keras_hub_export("keras_hub.models.DistilBertMaskedLM")
class DistilBertMaskedLM(MaskedLM):
    """An end-to-end DistilBERT model for the masked language modeling task.

    This model will train DistilBERT on a masked language modeling task.
    The model will predict labels for a number of masked tokens in the
    input data. For usage of this model with pre-trained weights, see the
    `from_preset()` constructor.

    This model can optionally be configured with a `preprocessor` layer, in
    which case inputs can be raw string features during `fit()`, `predict()`,
    and `evaluate()`. Inputs will be tokenized and dynamically masked during
    training and evaluation. This is done by default when creating the model
    with `from_preset()`.

    Disclaimer: Pre-trained models are provided on an "as is" basis, without
    warranties or conditions of any kind. The underlying model is provided by a
    third party and subject to a separate license, available
    [here](https://github.com/huggingface/transformers).

    Args:
        backbone: A `keras_hub.models.DistilBertBackbone` instance.
        preprocessor: A `keras_hub.models.DistilBertMaskedLMPreprocessor` or
            `None`. If `None`, this model will not apply preprocessing, and
            inputs should be preprocessed before calling the model.

    Examples:

    Raw string data.
    ```python
    features = ["The quick brown fox jumped.", "I forgot my homework."]

    # Pretrained language model.
    masked_lm = keras_hub.models.DistilBertMaskedLM.from_preset(
        "distil_bert_base_en_uncased",
    )
    masked_lm.fit(x=features, batch_size=2)

    # Re-compile (e.g., with a new learning rate).
    masked_lm.compile(
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        optimizer=keras.optimizers.Adam(5e-5),
        jit_compile=True,
    )
    # Access backbone programmatically (e.g., to change `trainable`).
    masked_lm.backbone.trainable = False
    # Fit again.
    masked_lm.fit(x=features, batch_size=2)
    ```

    Preprocessed integer data.
    ```python
    # Create preprocessed batch where 0 is the mask token.
    features = {
        "token_ids": np.array([[1, 2, 0, 4, 0, 6, 7, 8]] * 2),
        "padding_mask": np.array([[1, 1, 1, 1, 1, 1, 1, 1]] * 2),
        "mask_positions": np.array([[2, 4]] * 2)
    }
    # Labels are the original masked values.
    labels = [[3, 5]] * 2

    masked_lm = keras_hub.models.DistilBertMaskedLM.from_preset(
        "distil_bert_base_en_uncased",
        preprocessor=None,
    )
    masked_lm.fit(x=features, y=labels, batch_size=2)
    ```
    """

    backbone_cls = DistilBertBackbone
    preprocessor_cls = DistilBertMaskedLMPreprocessor

    def __init__(
        self,
        backbone,
        preprocessor=None,
        **kwargs,
    ):
        # === Layers ===
        self.backbone = backbone
        self.preprocessor = preprocessor
        self.masked_lm_head = MaskedLMHead(
            vocabulary_size=backbone.vocabulary_size,
            token_embedding=backbone.token_embedding,
            intermediate_activation="gelu",
            kernel_initializer=distilbert_kernel_initializer(),
            dtype=backbone.dtype_policy,
            name="mlm_head",
        )

        # === Functional Model ===
        inputs = {
            **backbone.input,
            "mask_positions": keras.Input(
                shape=(None,), dtype="int32", name="mask_positions"
            ),
        }
        backbone_outputs = backbone(backbone.input)
        outputs = self.masked_lm_head(
            backbone_outputs, inputs["mask_positions"]
        )
        super().__init__(
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )



import json
import os
import shutil
import warnings

import keras

from keras_hub.src.utils.transformers.export.gemma import get_gemma_config
from keras_hub.src.utils.transformers.export.gemma import get_gemma_weights_map

MODEL_CONFIGS = {
    "GemmaBackbone": get_gemma_config,
    # Add future models here, e.g., "LlamaBackbone": get_llama_config,
}

MODEL_EXPORTERS = {
    "GemmaBackbone": get_gemma_weights_map,
    # Add future models here, e.g., "LlamaBackbone": get_llama_weights_map,
}


def export_to_safetensors(keras_model, path):
    """Converts a Keras model to Hugging Face safetensor format.

    It does the following:
    - Extracts and maps weights from the Keras backbone to safetensors.
    - Saves the configuration as 'config.json'.
    - Saves weights in 'model.safetensors'.
    - Saves tokenizer assets.

    Args:
        keras_model: The Keras model to convert.
        path: str. Path of the directory to which the safetensors file,
          config and tokenizer will be saved.
    """
    backend = keras.config.backend()
    backbone = keras_model.backbone
    model_type = backbone.__class__.__name__

    if model_type not in MODEL_CONFIGS:
        raise ValueError(f"Config not implemented for {model_type}")

    if model_type not in MODEL_EXPORTERS:
        raise ValueError(f"Exporter not implemented for {model_type}")

    get_config_fn = MODEL_CONFIGS[model_type]
    hf_config = get_config_fn(backbone)

    get_weights_fn = MODEL_EXPORTERS[model_type]
    weights_dict = get_weights_fn(backbone)

    if not weights_dict:
        raise ValueError("No weights to save.")

    # Save config
    os.makedirs(path, exist_ok=True)
    config_path = os.path.join(path, "config.json")
    with open(config_path, "w") as f:
        json.dump(hf_config, f)

    # Save weights based on backend
    weights_path = os.path.join(path, "model.safetensors")
    if backend == "torch":
        from safetensors.torch import save_file

        weights_dict_contiguous = {
            k: v.value.contiguous() if hasattr(v, "value") else v.contiguous()
            for k, v in weights_dict.items()
        }
        save_file(
            weights_dict_contiguous, weights_path, metadata={"format": "pt"}
        )
    elif backend == "tensorflow":
        from safetensors.tensorflow import save_file

        save_file(weights_dict, weights_path, metadata={"format": "pt"})
    elif backend == "jax":
        from safetensors.flax import save_file

        save_file(weights_dict, weights_path, metadata={"format": "pt"})
    else:
        raise ValueError(f"Unsupported backend: {backend}")

    # Save tokenizer assets
    keras_model.preprocessor.tokenizer.save_assets(path)

    # Rename vocabulary file
    vocab_spm_path = os.path.join(path, "vocabulary.spm")
    tokenizer_model_path = os.path.join(path, "tokenizer.model")
    if os.path.exists(vocab_spm_path):
        shutil.move(vocab_spm_path, tokenizer_model_path)
    else:
        warnings.warn(
            f"{vocab_spm_path} not found. Tokenizer may not load "
            "correctly. Ensure that the tokenizer configuration "
            "is correct and that the vocabulary file is present "
            "in the original model."
        )



import numpy as np
from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.causal_lm import CausalLM
from keras_hub.src.models.gemma3.gemma3_backbone import Gemma3Backbone
from keras_hub.src.models.gemma3.gemma3_causal_lm_preprocessor import (
    Gemma3CausalLMPreprocessor,
)
from keras_hub.src.utils.tensor_utils import any_equal

try:
    import tensorflow as tf
except ImportError:
    tf = None


@keras_hub_export("keras_hub.models.Gemma3CausalLM")
class Gemma3CausalLM(CausalLM):
    """An end-to-end multimodal Gemma3 model for causal language modeling.

    A causal language model (LM) predicts the next token based on previous
    tokens. This task setup can be used to train the model unsupervised on
    images and plain text inputs, or to autoregressively generate plain text
    similar to the data used for training. Note that the model is
    image-text in, text out.

    This model has a `generate()` method, which generates text based on a
    prompt. The generation strategy used is controlled by an additional
    `sampler` argument on `compile()`. You can recompile the model with
    different `keras_hub.samplers` objects to control the generation. By
    default, `"greedy"` sampling will be used.

    This model can optionally be configured with a `preprocessor` layer, in
    which case it will automatically apply preprocessing to string inputs during
    `fit()`, `predict()`, `evaluate()` and `generate()`. This is done by default
    when creating the model with `from_preset()`.

    Args:
        preprocessor: A `keras_hub.models.Gemma3CausalLMPreprocessor` or
            `None`. If `None`, this model will not apply preprocessing, and
            inputs should be preprocessed before calling the model.
        backbone: A `keras_hub.models.Gemma3Backbone` instance.
    """

    backbone_cls = Gemma3Backbone
    preprocessor_cls = Gemma3CausalLMPreprocessor

    def __init__(
        self,
        preprocessor,
        backbone,
        **kwargs,
    ):
        # === Layers ===
        self.preprocessor = preprocessor
        self.backbone = backbone

        # === Functional Model ===
        # This must be "backbone.input" i.e. the full input structure,
        # rather than "backbone.inputs" which is the flattened list of inputs.
        inputs = backbone.input
        hidden_state = backbone(inputs=inputs)
        outputs = backbone.token_embedding(hidden_state, reverse=True)

        super().__init__(
            inputs=inputs,
            outputs=outputs,
            **kwargs,
        )

    def compile(
        self,
        optimizer="auto",
        loss="auto",
        *,
        weighted_metrics="auto",
        sampler="greedy",
        **kwargs,
    ):
        super().compile(
            optimizer=optimizer,
            loss=loss,
            weighted_metrics=weighted_metrics,
            sampler=sampler,
            **kwargs,
        )

    def _normalize_generate_inputs(
        self,
        inputs,
    ):
        """Overrides the superclass' method to handle unbatched image inputs."""
        if tf and isinstance(inputs, tf.data.Dataset):
            return inputs.as_numpy_iterator(), False

        if self.preprocessor is None:
            return [inputs], False

        def normalize(x):
            if isinstance(x, str):
                return [x], True
            if tf and isinstance(x, tf.Tensor) and x.shape.rank == 0:
                return x[tf.newaxis], True
            return x, False

        if isinstance(inputs, dict):
            inputs["prompts"], input_is_scalar = normalize(inputs["prompts"])

            # If prompt is scalar, images can be either a 3D NumPy array/Tensor,
            # or, list of 3D NumPy arrays. Let's uprank images.
            if input_is_scalar and "images" in inputs:
                x = inputs["images"]
                if isinstance(x, np.ndarray) and len(x.shape) == 3:
                    inputs["images"] = [x]
                elif tf and isinstance(x, tf.Tensor) and x.shape.rank == 3:
                    inputs["images"] = x[tf.newaxis]
                elif isinstance(x, list):
                    inputs["images"] = [x]

            if "responses" in inputs:
                inputs["responses"], _ = normalize(inputs["responses"])
        else:
            inputs, input_is_scalar = normalize(inputs)

        return [inputs], input_is_scalar

    def call_with_cache(
        self,
        token_ids,
        cache,
        cache_update_index,
        img_embeddings=None,
        vision_mask=None,
        padding_mask=None,
        vision_indices=None,
        cache_update_mask=None,
    ):
        """Forward pass of `Gemma3CausalLM` with cache.

        `call_with_cache` adds an additional forward pass for the model for
        autoregressive inference. Unlike calling the model directly, this method
        allows caching previous key/value Tensors in multi-head attention layer,
        and avoids recomputing the outputs of seen tokens.

        Args:
            token_ids: a dense int Tensor with shape `(batch_size, max_length)`.
            cache: a dense float Tensor, the cache of key and value.
            cache_update_index: int, or int Tensor. The index of current inputs
                in the whole sequence.
            img_embeddings: a dense float Tensor with shape
                `(batch_size, image_sequence_length, hidden_dim)`.
            padding_mask: a dense int Tensor with shape
                `(batch_size, max_length)`.

        Returns:
            A (logits, hidden_states, cache) tuple. Where `logits` is the
            language model logits for the input token_ids, `hidden_states` is
            the final hidden representation of the input tokens, and `cache` is
            the decoding cache.
        """

        text_embeddings = self.backbone.token_embedding(token_ids)
        text_embeddings = text_embeddings * ops.cast(
            ops.sqrt(self.backbone.hidden_dim), text_embeddings.dtype
        )

        # Interleaving logic.
        ## == Interleaving text and images ==
        # Place the image embeddings in the right position in `text_embeddings`.
        if img_embeddings is not None:
            x = self.backbone.interleave_embeddings(
                image_embeddings=img_embeddings,
                text_embeddings=text_embeddings,
                vision_indices=vision_indices,
            )
        else:
            x = text_embeddings

        # Each decoder layer has a cache; we update them separately.
        caches = []
        for i, transformer_layer in enumerate(self.backbone.transformer_layers):
            current_cache = cache[:, i, ...]
            x, next_cache = transformer_layer(
                x,
                cache=current_cache,
                cache_update_index=cache_update_index,
                padding_mask=padding_mask,
                vision_mask=vision_mask,
                cache_update_mask=cache_update_mask,
            )
            caches.append(next_cache)
        cache = ops.stack(caches, axis=1)
        hidden_states = x = self.backbone.layer_norm(x)
        logits = self.backbone.token_embedding(x, reverse=True)
        return logits, hidden_states, cache

    def _build_cache(
        self,
        token_ids,
        img_embeddings,
        vision_mask,
        padding_mask,
        vision_indices,
    ):
        """Build an empty cache for use with `call_with_cache()`."""
        batch_size = ops.shape(token_ids)[0]
        max_length = (
            ops.shape(token_ids)[1]
            # + self.backbone.image_sequence_length
        )
        num_layers = self.backbone.num_layers
        num_heads = self.backbone.num_key_value_heads
        head_dim = self.backbone.head_dim
        shape = [batch_size, num_layers, 2, max_length, num_heads, head_dim]
        cache = ops.zeros(shape, dtype=self.compute_dtype)
        # Seed the cache.
        logits, hidden_states, cache = self.call_with_cache(
            token_ids=token_ids,
            img_embeddings=img_embeddings,
            vision_mask=vision_mask,
            cache=cache,
            cache_update_index=0,
            padding_mask=padding_mask,
            vision_indices=vision_indices,
            cache_update_mask=None,
        )
        return hidden_states, cache

    def generate_step(self, inputs, stop_token_ids=[106]):
        """A compilable generation function for a single batch of inputs.

        This function represents the inner, XLA-compilable, generation function
        for a single batch of inputs. Inputs should have the same structure as
        model inputs, a dictionary with keys `"token_ids"` and `"padding_mask"`.

        Args:
            inputs: A dictionary with two keys `"token_ids"` and
                `"padding_mask"` and batched tensor values.
            stop_token_ids: Tuple of id's of end token's to stop on. If all
                sequences have produced a new stop token, generation
                will stop.
        """

        token_ids, padding_mask, images, vision_mask, vision_indices = (
            inputs["token_ids"],
            inputs["padding_mask"],
            inputs.get("images", None),
            inputs.get("vision_mask", None),
            inputs.get("vision_indices", None),
        )
        if not self.backbone.text_only_model:
            # Handle an unbatched image. Unlike `token_ids` and
            # `padding_mask`, this will not automatically be upranked.
            if len(ops.shape(images)) == 4:
                images = ops.expand_dims(images, axis=0)
            if len(ops.shape(vision_mask)) == 1:
                vision_mask = ops.expand_dims(vision_mask, axis=0)
            if len(ops.shape(vision_indices)) == 1:
                vision_indices = ops.expand_dims(vision_indices, axis=0)
            img_embeddings = self.backbone.vision_encoder(images)
        else:
            img_embeddings = None
            vision_mask = None
            vision_indices = None

        # Create and seed cache with a single forward pass.
        hidden_states, cache = self._build_cache(
            token_ids,
            img_embeddings,
            vision_mask,
            padding_mask,
            vision_indices,
        )

        # Compute the lengths of all user inputted tokens ids.
        row_lengths = ops.sum(ops.cast(padding_mask, "int32"), axis=-1)
        # Start at the first index that has no user inputted id.
        index = ops.min(row_lengths)

        def next(prompt, cache, index):
            # The cache index is the index of our previous token.
            cache_update_index = index - 1
            batch_size = ops.shape(prompt)[0]
            prompt = ops.slice(prompt, [0, index - 1], [batch_size, 1])
            sliced_cache_update_mask = ops.slice(
                ~padding_mask, [0, index - 1], [batch_size, 1]
            )
            logits, hidden_states, cache = self.call_with_cache(
                token_ids=prompt,
                cache=cache,
                cache_update_index=cache_update_index,
                cache_update_mask=sliced_cache_update_mask,
            )
            return (
                ops.squeeze(logits, axis=1),
                ops.squeeze(hidden_states, axis=1),
                cache,
            )

        token_ids = self.sampler(
            next=next,
            prompt=token_ids,
            cache=cache,
            index=index,
            mask=padding_mask,
            stop_token_ids=stop_token_ids,
            hidden_states=hidden_states,
            model=self,
        )

        # Compute an output padding mask with the token ids we updated.
        if stop_token_ids is not None:
            # Build a mask of `stop_token_ids` locations not in the original
            # prompt (not in locations where `padding_mask` is True).
            end_locations = any_equal(
                token_ids, stop_token_ids, ops.logical_not(padding_mask)
            )

            end_locations = ops.cast(end_locations, "int32")
            # Use cumsum to get ones in all locations after end_locations.
            cumsum = ops.cast(ops.cumsum(end_locations, axis=-1), "int32")
            overflow = cumsum - end_locations
            # Our padding mask is the inverse of these overflow locations.
            padding_mask = ops.logical_not(ops.cast(overflow, "bool"))
        else:
            # Without early stopping, all locations will have been updated.
            padding_mask = ops.ones_like(token_ids, dtype="bool")
        return {
            "token_ids": token_ids,
            "padding_mask": padding_mask,
            "images": images,
        }

    def generate(
        self,
        inputs,
        max_length=None,
        stop_token_ids="auto",
        strip_prompt=False,
    ):
        # If `auto`, add `<end_of_turn>` as a stop token too.
        if self.preprocessor is None and stop_token_ids == "auto":
            raise ValueError(
                "A `preprocessor` must be attached to the model if "
                '`stop_token_ids="auto"`. Currently `preprocessor=None`. To '
                "call `generate()` with preprocessing detached, either pass "
                "`stop_token_ids=None` to always generate until `max_length` "
                "or pass a tuple of token ids that should terminate generation "
                "as `stop_token_ids`."
            )
        elif stop_token_ids == "auto":
            stop_token_ids = [
                self.preprocessor.tokenizer.end_token_id,
                self.preprocessor.tokenizer.token_to_id("<end_of_turn>"),
            ]

        return super().generate(
            inputs,
            max_length=max_length,
            stop_token_ids=stop_token_ids,
            strip_prompt=strip_prompt,
        )



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.image_converter import ImageConverter
from keras_hub.src.models.retinanet.retinanet_backbone import RetinaNetBackbone


@keras_hub_export("keras_hub.layers.RetinaNetImageConverter")
class RetinaNetImageConverter(ImageConverter):
    backbone_cls = RetinaNetBackbone



import keras.ops as ops


def get_gemma_config(backbone):
    hf_config = {
        "vocab_size": backbone.vocabulary_size,
        "num_hidden_layers": backbone.num_layers,
        "num_attention_heads": backbone.num_query_heads,
        "num_key_value_heads": backbone.num_key_value_heads,
        "hidden_size": backbone.hidden_dim,
        "intermediate_size": backbone.intermediate_dim // 2,
        "head_dim": backbone.head_dim,
        "max_position_embeddings": 8192,
    }
    return hf_config


def get_gemma_weights_map(backbone):
    weights_dict = {}

    # Map token embedding
    token_embedding_layer = backbone.get_layer("token_embedding")
    weights_dict["model.embed_tokens.weight"] = token_embedding_layer.weights[0]

    for i in range(backbone.num_layers):
        decoder_layer = backbone.get_layer(f"decoder_block_{i}")

        # Pre-attention normalization
        weights_dict[f"model.layers.{i}.input_layernorm.weight"] = (
            decoder_layer.pre_attention_norm.weights[0]
        )

        # Attention query projection
        query_kernel = decoder_layer.attention.query_dense.weights[0]
        query_kernel = ops.transpose(query_kernel, axes=(1, 0, 2))
        query_kernel = ops.reshape(query_kernel, (-1, backbone.hidden_dim))
        query_kernel = ops.transpose(query_kernel)
        weights_dict[f"model.layers.{i}.self_attn.q_proj.weight"] = query_kernel

        # Attention key projection
        key_kernel = decoder_layer.attention.key_dense.weights[0][0]
        weights_dict[f"model.layers.{i}.self_attn.k_proj.weight"] = (
            ops.transpose(key_kernel)
        )

        # Attention value projection
        value_kernel = decoder_layer.attention.value_dense.weights[0][0]
        weights_dict[f"model.layers.{i}.self_attn.v_proj.weight"] = (
            ops.transpose(value_kernel)
        )

        # Attention output projection
        out_kernel = decoder_layer.attention.output_dense.weights[0]
        out_kernel = ops.transpose(out_kernel, axes=(2, 0, 1))
        out_kernel = ops.reshape(out_kernel, (backbone.hidden_dim, -1))
        weights_dict[f"model.layers.{i}.self_attn.o_proj.weight"] = out_kernel

        # Post-attention normalization
        weights_dict[f"model.layers.{i}.post_attention_layernorm.weight"] = (
            decoder_layer.pre_ffw_norm.weights[0]
        )

        # MLP gate projection
        gate_kernel = decoder_layer.gating_ffw.weights[0]
        weights_dict[f"model.layers.{i}.mlp.gate_proj.weight"] = ops.transpose(
            gate_kernel
        )

        # MLP up projection
        up_kernel = decoder_layer.gating_ffw_2.weights[0]
        weights_dict[f"model.layers.{i}.mlp.up_proj.weight"] = ops.transpose(
            up_kernel
        )

        # MLP down projection
        down_kernel = decoder_layer.ffw_linear.weights[0]
        weights_dict[f"model.layers.{i}.mlp.down_proj.weight"] = ops.transpose(
            down_kernel
        )

    # Map final normalization
    weights_dict["model.norm.weight"] = backbone.get_layer(
        "final_normalization"
    ).weights[0]

    # Tie weights, but clone to avoid sharing memory issues
    weights_dict["lm_head.weight"] = ops.copy(token_embedding_layer.weights[0])

    return weights_dict



import keras

BN_AXIS = 3


class FusedMBConvBlock(keras.layers.Layer):
    """Implementation of the FusedMBConv block

    Also known as a Fused Mobile Inverted Residual Bottleneck block from:
    [EfficientNet-EdgeTPU](https://ai.googleblog.com/2019/08/efficientnet-edgetpu-creating.html)
    [EfficientNetV2: Smaller Models and Faster Training](https://arxiv.org/abs/2104.00298v3).

    FusedMBConv blocks are based on MBConv blocks, and replace the depthwise and
    1x1 output convolution blocks with a single 3x3 convolution block, fusing
    them together - hence the name "FusedMBConv". Alongside MBConv blocks, they
    can be used in mobile-oriented and efficient architectures, and are present
    in architectures EfficientNet.

    FusedMBConv blocks follow a narrow-wide-narrow structure - expanding a 1x1
    convolution, performing Squeeze-Excitation and then applying a 3x3
    convolution, which is a more efficient operation than conventional
    wide-narrow-wide structures.

    As they're frequently used for models to be deployed to edge devices,
    they're implemented as a layer for ease of use and re-use.

    Args:
        input_filters: int, the number of input filters
        output_filters: int, the number of output filters
        expand_ratio: default 1, the ratio by which input_filters are multiplied
            to expand the structure in the middle expansion phase
        kernel_size: default 3, the kernel_size to apply to the expansion phase
            convolutions
        strides: default 1, the strides to apply to the expansion phase
            convolutions
        data_format: str, channels_last (default) or channels_first, expects
            tensors to be of shape (N, H, W, C) or (N, C, H, W) respectively
        se_ratio: default 0.0, The filters used in the Squeeze-Excitation phase,
            and are chosen as the maximum between 1 and input_filters*se_ratio
        batch_norm_momentum: default 0.9, the BatchNormalization momentum
        batch_norm_epsilon: default 1e-3, float, epsilon for batch norm
            calcualtions. Used in denominator for calculations to prevent divide
            by 0 errors.
        activation: default "swish", the activation function used between
            convolution operations
        projection_activation: default None, the activation function to use
            after the output projection convoultion
        dropout: float, the optional dropout rate to apply before the output
            convolution, defaults to 0.2
        nores: bool, default False, forces no residual connection if True,
            otherwise allows it if False.
        projection_kernel_size: default 1, the kernel_size to apply to the
            output projection phase convolution

    Returns:
        A tensor representing a feature map, passed through the FusedMBConv
        block

    Note:
        Not intended to be used outside of the EfficientNet architecture.
    """

    def __init__(
        self,
        input_filters,
        output_filters,
        expand_ratio=1,
        kernel_size=3,
        strides=1,
        data_format="channels_last",
        se_ratio=0.0,
        batch_norm_momentum=0.9,
        batch_norm_epsilon=1e-3,
        activation="swish",
        projection_activation=None,
        dropout=0.2,
        nores=False,
        projection_kernel_size=1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.input_filters = input_filters
        self.output_filters = output_filters
        self.expand_ratio = expand_ratio
        self.kernel_size = kernel_size
        self.strides = strides
        self.data_format = data_format
        self.se_ratio = se_ratio
        self.batch_norm_momentum = batch_norm_momentum
        self.batch_norm_epsilon = batch_norm_epsilon
        self.activation = activation
        self.projection_activation = projection_activation
        self.dropout = dropout
        self.nores = nores
        self.projection_kernel_size = projection_kernel_size
        self.filters = self.input_filters * self.expand_ratio
        self.filters_se = max(1, int(input_filters * se_ratio))

        padding_pixels = kernel_size // 2
        self.conv1_pad = keras.layers.ZeroPadding2D(
            padding=(padding_pixels, padding_pixels),
            name=self.name + "expand_conv_pad",
        )
        self.conv1 = keras.layers.Conv2D(
            filters=self.filters,
            kernel_size=kernel_size,
            strides=strides,
            kernel_initializer=self._conv_kernel_initializer(),
            padding="valid",
            data_format=data_format,
            use_bias=False,
            name=self.name + "expand_conv",
        )
        self.bn1 = keras.layers.BatchNormalization(
            axis=BN_AXIS,
            momentum=self.batch_norm_momentum,
            epsilon=self.batch_norm_epsilon,
            name=self.name + "expand_bn",
        )
        self.act = keras.layers.Activation(
            self.activation, name=self.name + "expand_activation"
        )

        self.se_conv1 = keras.layers.Conv2D(
            self.filters_se,
            1,
            padding="same",
            data_format=data_format,
            activation=self.activation,
            kernel_initializer=self._conv_kernel_initializer(),
            name=self.name + "se_reduce",
        )

        self.se_conv2 = keras.layers.Conv2D(
            self.filters,
            1,
            padding="same",
            data_format=data_format,
            activation="sigmoid",
            kernel_initializer=self._conv_kernel_initializer(),
            name=self.name + "se_expand",
        )

        padding_pixels = projection_kernel_size // 2
        self.output_conv_pad = keras.layers.ZeroPadding2D(
            padding=(padding_pixels, padding_pixels),
            name=self.name + "project_conv_pad",
        )
        self.output_conv = keras.layers.Conv2D(
            filters=self.output_filters,
            kernel_size=projection_kernel_size,
            strides=1,
            kernel_initializer=self._conv_kernel_initializer(),
            padding="valid",
            data_format=data_format,
            use_bias=False,
            name=self.name + "project_conv",
        )

        self.bn2 = keras.layers.BatchNormalization(
            axis=BN_AXIS,
            momentum=self.batch_norm_momentum,
            epsilon=self.batch_norm_epsilon,
            name=self.name + "project_bn",
        )

        if self.projection_activation:
            self.projection_act = keras.layers.Activation(
                self.projection_activation, name=self.name + "projection_act"
            )

        if self.dropout:
            self.dropout_layer = keras.layers.Dropout(
                self.dropout,
                noise_shape=(None, 1, 1, 1),
                name=self.name + "drop",
            )

    def _conv_kernel_initializer(
        self,
        scale=2.0,
        mode="fan_out",
        distribution="truncated_normal",
        seed=None,
    ):
        return keras.initializers.VarianceScaling(
            scale=scale, mode=mode, distribution=distribution, seed=seed
        )

    def build(self, input_shape):
        if self.name is None:
            self.name = keras.backend.get_uid("block0")

    def call(self, inputs):
        # Expansion phase
        x = self.conv1_pad(inputs)
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.act(x)

        # Squeeze and excite
        if 0 < self.se_ratio <= 1:
            se = keras.layers.GlobalAveragePooling2D(
                name=self.name + "se_squeeze",
                data_format=self.data_format,
            )(x)
            if BN_AXIS == 1:
                se_shape = (self.filters, 1, 1)
            else:
                se_shape = (1, 1, self.filters)

            se = keras.layers.Reshape(se_shape, name=self.name + "se_reshape")(
                se
            )

            se = self.se_conv1(se)
            se = self.se_conv2(se)

            x = keras.layers.multiply([x, se], name=self.name + "se_excite")

        # Output phase:
        x = self.output_conv_pad(x)
        x = self.output_conv(x)
        x = self.bn2(x)
        if self.expand_ratio == 1 and self.projection_activation:
            x = self.projection_act(x)

        # Residual:
        if (
            self.strides == 1
            and self.input_filters == self.output_filters
            and not self.nores
        ):
            if self.dropout:
                x = self.dropout_layer(x)
            x = keras.layers.Add(name=self.name + "add")([x, inputs])
        return x

    def get_config(self):
        config = {
            "input_filters": self.input_filters,
            "output_filters": self.output_filters,
            "expand_ratio": self.expand_ratio,
            "kernel_size": self.kernel_size,
            "strides": self.strides,
            "data_format": self.data_format,
            "se_ratio": self.se_ratio,
            "batch_norm_momentum": self.batch_norm_momentum,
            "batch_norm_epsilon": self.batch_norm_epsilon,
            "activation": self.activation,
            "projection_activation": self.projection_activation,
            "dropout": self.dropout,
            "nores": self.nores,
            "projection_kernel_size": self.projection_kernel_size,
        }

        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))
